using System.Text.Json;
using System.Text.Json.Serialization;

namespace View_by_Distance.Shared.Models;

public class PersonContainer : Properties.IPersonContainer
{

    public int? ApproximateYears { init; get; }
    public Person? Person { init; get; }
    public PersonBirthday[]? Birthdays { init; get; }
    public string[] DisplayDirectoryAllFiles { init; get; }
    public string DisplayDirectoryName { init; get; }
    public long? Key { init; get; }

    [JsonConstructor]
    public PersonContainer(int? approximateYears, Person? person, PersonBirthday[]? personBirthdays, string[] personDisplayDirectoryAllFiles, string personDisplayDirectoryName, long? personKey)
    {
        ApproximateYears = approximateYears;
        Person = person;
        Birthdays = personBirthdays;
        DisplayDirectoryAllFiles = personDisplayDirectoryAllFiles;
        DisplayDirectoryName = personDisplayDirectoryName;
        Key = personKey;
    }

    public PersonContainer(int? approximateYears, string[] personDisplayDirectoryAllFiles, string personDisplayDirectoryName) :
        this(approximateYears, null, null, personDisplayDirectoryAllFiles, personDisplayDirectoryName, null)
    { }

    public PersonContainer(int? approximateYears, PersonBirthday personBirthday, string personDisplayDirectoryName, long personKey) :
        this(approximateYears, null, new PersonBirthday[] { personBirthday }, Array.Empty<string>(), personDisplayDirectoryName, personKey)
    { }

    public PersonContainer(PersonBirthday personBirthday, string personDisplayDirectoryName) :
        this(Stateless.Methods.IAge.GetApproximateYears(personDisplayDirectoryName), Stateless.Methods.IPerson.GetPerson(personDisplayDirectoryName, personBirthday.Value.Ticks, personBirthday), new PersonBirthday[] { personBirthday }, Array.Empty<string>(), personDisplayDirectoryName, personBirthday.Value.Ticks)
    { }

    public override string ToString()
    {
        string result = JsonSerializer.Serialize(this, new JsonSerializerOptions() { WriteIndented = true });
        return result;
    }

}